from meross_iot.cloud.abilities import *
from meross_iot.cloud.device import AbstractMerossDevice
from meross_iot.logger import POWER_PLUGS_LOGGER as l
from meross_iot.meross_event import DeviceHubStatusEvent
from meross_iot.cloud.devices.subdevices.generic import GenericSubDevice
from meross_iot.cloud.devices.subdevices.thermostats import ValveSubDevice


def get_sub_device(subdevice_id, hub, **client_data):
    if 'mts100v3' in client_data or 'mts100' in client_data:
        return ValveSubDevice(subdevice_id=subdevice_id, hub=hub, **client_data)
    else:
        # TODO: log unknown subdevice
        return GenericSubDevice(subdevice_id=subdevice_id, hub=hub, **client_data)


class GenericHub(AbstractMerossDevice):
    # Handles the state of this specific HUB
    _state = {}
    _sub_devices = {}

    def __init__(self, cloud_client, device_uuid, **kwords):
        super(GenericHub, self).__init__(cloud_client, device_uuid, **kwords)

    def _update_subdevice_data(self, client_data, namespace):
        client_id = client_data.get('id')

        # Check if the sensor is already present into the list of subdevices
        subdev = self._sub_devices.get(client_id)
        if subdev is None:
            subdev = get_sub_device(subdevice_id=client_id, hub=self, **client_data)
            self._sub_devices[client_id] = subdev

            # Trigger specific status update
            subdev.update_all()

        subdev.handle_push_event(client_data, namespace)

    def _get_status_impl(self):
        res = {}
        data = self.get_sys_data()['all']
        hub_data = data.get('digest').get('hub')
        res['hub_id'] = hub_data.get('hubId')
        res['mode'] = hub_data.get('mode')
        for subdevice in hub_data.get('subdevice'):
            self._update_subdevice_data(subdevice, None)
        return res

    def get_status(self):
        with self._state_lock:
            if self._state == {}:
                self._state = self._get_status_impl()
            return self._state

    def _handle_push_notification(self, namespace, payload, from_myself=False):
        def fire_hub_state_change(dev, subdevice_id, state_data, f_myself):
            evt = DeviceHubStatusEvent(hubdevice=dev, subdevice_id=subdevice_id, state=state_data,
                                       generated_by_myself=f_myself)
            self.fire_event(evt)

        def update_and_fire_event(subdevice_data, namespace):
            subdevice_id = subdevice_data.get('id')
            self._update_subdevice_data(subdevice_data, namespace)
            fire_hub_state_change(self, subdevice_id, self.get_subdevice_state(subdevice_id), from_myself)

        with self._state_lock:
            if namespace == HUB_TOGGLEX:
                for sensor_data in payload['togglex']:
                    update_and_fire_event(sensor_data, namespace)

            elif namespace == REPORT:
                pass

            elif namespace == HUB_MTS100_MODE:
                for sensor_data in payload['mode']:
                    update_and_fire_event(sensor_data, namespace)

            # Target temperature set on the device
            elif namespace == HUB_MTS100_TEMPERATURE:
                for sensor_data in payload['temperature']:
                    update_and_fire_event(sensor_data, namespace)

            else:
                l.warn("Unknown/Unsupported namespace/command: %s" % namespace)

    def _togglex(self, subdevice_id, status, channel=0, callback=None):
        payload = {'togglex': [{'id': subdevice_id, "onoff": status, "channel": channel}]}
        return self.execute_command('SET', HUB_TOGGLEX, payload, callback=callback)

    def get_subdevice_state(self, subdevice_id):
        return self._sub_devices.get(subdevice_id)

    def get_subdevices(self):
        return self._sub_devices.keys()

    def turn_on_subdevice(self, subdevice_id, channel=0, callback=None):
        return self._togglex(subdevice_id, 1, channel, callback=callback)

    def turn_off_subdevice(self, subdevice_id, channel=0, callback=None):
        return self._togglex(subdevice_id, 0, channel, callback=callback)

    def __str__(self):
        self.get_status()
        base_str = super().__str__()
        with self._state_lock:
            if not self.online:
                return base_str
            # TODO: fix this method. We'd probably want to print some more meaningful info
            return "%s [ %s ]" % (base_str, ",".join(self.get_subdevices()))



import base64
import hashlib
import json
import random
import string
import time
import requests
from meross_iot.credentials import MerossCloudCreds
from meross_iot.logger import API_LOGGER as l

# Appears to be used as a part of the signature algorithm as constant "salt" (kinda useless)
_SECRET = "23x17ahWarFH6w29"
_MEROSS_URL = "https://iot.meross.com"
_LOGIN_URL = "%s%s" % (_MEROSS_URL, "/v1/Auth/Login")
_LOG_URL = "%s%s" % (_MEROSS_URL, "/v1/log/user")
_DEV_LIST = "%s%s" % (_MEROSS_URL, "/v1/Device/devList")
_HUB_DUBDEV_LIST = "%s%s" % (_MEROSS_URL, "/v1/Hub/getSubDevices")


class MerossHttpClient:
    _cloud_creds = None

    _email = None
    _password = None
    _authenticated = False

    def __init__(self, email, password):
        self._email = email
        self._password = password

    def _authenticated_post(self,
                            url,  # type: str
                            params_data  # type: dict
                            ):

        nonce = self._generate_nonce(16)
        timestamp_millis = int(round(time.time() * 1000))
        login_params = self._encode_params(params_data)

        # Generate the md5-hash (called signature)
        m = hashlib.md5()
        datatosign = '%s%s%s%s' % (_SECRET, timestamp_millis, nonce, login_params)
        m.update(datatosign.encode("utf8"))
        md5hash = m.hexdigest()

        headers = {
            "Authorization": "Basic" if self._cloud_creds is None else "Basic %s" % self._cloud_creds.token,
            "vender": "Meross",
            "AppVersion": "1.3.0",
            "AppLanguage": "EN",
            "User-Agent": "okhttp/3.6.0"
        }

        payload = {
            'params': login_params,
            'sign': md5hash,
            'timestamp': timestamp_millis,
            'nonce': nonce
        }

        # Perform the request.
        r = requests.post(url, data=payload, headers=headers)

        # Check if that is ok.
        if r.status_code != 200:
            raise AuthenticatedPostException()

        # Save returned value
        jsondata = r.json()
        # print(jsondata)

        if jsondata["info"].lower() != "success":
            raise AuthenticatedPostException()

        return jsondata["data"]

    @staticmethod
    def _encode_params(parameters  # type: dict
                       ):
        jsonstring = json.dumps(parameters)
        return str(base64.b64encode(jsonstring.encode("utf8")), "utf8")

    @staticmethod
    def _generate_nonce(length):
        return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(length))

    def _login(self):
        try:
            data = {"email": self._email, "password": self._password}
            response_data = self._authenticated_post(_LOGIN_URL, params_data=data)
            creds = MerossCloudCreds()
            creds.token = response_data["token"]
            creds.key = response_data["key"]
            creds.user_id = response_data["userid"]
            creds.user_email = response_data["email"]
            self._cloud_creds = creds
            self._authenticated = True
        except:
            l.exception("Error occurred when logging in.")
            return False

        try:
            # The protocol does not really need the following call. However we want to be nice do it anyways
            self._log()
        except:
            l.exception("Error occurred when logging in.")
            pass

        return True

    def _log(self):
        data = {'extra': {}, 'model': 'Android,Android SDK built for x86_64', 'system': 'Android',
                'uuid': '493dd9174941ed58waitForOpenWifi', 'vendor': 'Meross', 'version': '6.0'}
        self._authenticated_post(_LOG_URL, params_data=data)

    def list_devices(self):
        if not self._authenticated and not self._login():
            raise UnauthorizedException()

        return self._authenticated_post(_DEV_LIST, {})

    def get_cloud_credentials(self):
        if not self._authenticated and not self._login():
            raise UnauthorizedException()

        return self._cloud_creds

    def list_hub_subdevices(self, hub_id):
        if not self._authenticated and not self._login():
            raise UnauthorizedException()

        return self._authenticated_post(_HUB_DUBDEV_LIST, {"uuid": hub_id})


class AuthenticatedPostException(Exception):
    pass


class UnauthorizedException(Exception):
    pass



from meross_iot.cloud.abilities import HUB_TOGGLEX
from meross_iot.logger import SUBDEVICE_LOGGER as l


class GenericSubDevice:
    id = None
    _hub = None
    online = False
    onoff = None
    last_active_time = None

    def __init__(self, subdevice_id, hub, **kwords):
        self.id = subdevice_id
        self._hub = hub
        self.update_state(kwords)

    def update_state(self, subdevice_data):
        # Update the sensor data
        for k, v in subdevice_data.items():
            if k == 'id':
                continue
            elif k == 'status':
                self.online = v == 1
            elif k == 'onoff':
                self.onoff = v
            elif k == 'lastActiveTime':
                self.last_active_time = v

    def handle_push_event(self, subdevice_data, namespace):
        if namespace == HUB_TOGGLEX:
            self.onoff = subdevice_data.get('onoff')
        elif namespace is None:
            pass
        else:
            l.warn("Unsupported namespace push event: %s" % namespace)

    def update_all(self):
        pass

    def __str__(self):
        return 'ID: %s' % self.id

    # TODO
    # def __str__(self):
    #    pass



import datetime
from threading import RLock, Condition

from meross_iot.cloud.client_status import ClientStatus
from meross_iot.cloud.exceptions.StatusTimeoutException import StatusTimeoutException
from meross_iot.cloud.timeouts import SHORT_TIMEOUT
from meross_iot.meross_event import ClientConnectionEvent
from meross_iot.logger import CONNECTION_MANAGER_LOGGER as l


class ConnectionStatusManager(object):
    # The connection status of the device is represented by the following variable.
    # It is protected by the following variable, called _client_connection_status_lock.
    # The child classes should never change/access these variables directly, though.
    _status = None
    _lock = None

    # List of callbacks that should be called when an event occurs
    _connection_event_callbacks = None
    _connection_event_callbacks_lock = None

    # This condition object is used to synchronize multiple threads waiting for the connection to
    # get into a specific state.
    _status_condition = None

    def __init__(self):
        self._connection_event_callbacks_lock = RLock()
        self._connection_event_callbacks = []

        self._lock = RLock()
        self._status_condition = Condition(self._lock)
        self._status = ClientStatus.INITIALIZED

    def get_status(self):
        with self._status_condition:
            return self._status

    def update_status(self, status):
        old_status = None
        new_status = None
        with self._status_condition:
            old_status = self._status
            new_status = status
            self._status = status
            self._status_condition.notify_all()

        # If the connection status has changed, fire the event.
        if old_status != new_status:
            self._fire_connection_event(new_status)

    def check_status(self, expected_status):
        with self._lock:
            return expected_status == self._status

    def check_status_in(self, expected_statuses):
        with self._lock:
            return self._status in expected_statuses

    def wait_for_status(self, expected_status, timeout=SHORT_TIMEOUT):
        start_time = datetime.datetime.now()
        with self._status_condition:
            while self._status != expected_status:
                elapsed = datetime.datetime.now().timestamp() - start_time.timestamp()
                to = timeout - elapsed
                timeout_hit = to < 0 or not self._status_condition.wait(to)

                if timeout_hit:
                    # An error has occurred
                    raise StatusTimeoutException("Error while waiting for status %s. Last status is: %s" %
                                                 (expected_status, self._status))

    # ------------------------------------------------------------------------------------------------
    # Event Handling
    # ------------------------------------------------------------------------------------------------
    def register_connection_event_callback(self, callback):
        with self._connection_event_callbacks_lock:
            if callback not in self._connection_event_callbacks:
                self._connection_event_callbacks.append(callback)
            else:
                l.debug("Callback was already registered.")

    def unregister_connection_event_callback(self, callback):
        with self._connection_event_callbacks_lock:
            if callback in self._connection_event_callbacks:
                self._connection_event_callbacks.remove(callback)
            else:
                l.debug("Callback was present: nothing to unregister.")

    def _fire_connection_event(self, connection_status):
        evt = ClientConnectionEvent(current_status=connection_status)
        for c in self._connection_event_callbacks:
            try:
                c(evt)
            except:
                l.exception("Unhandled error occurred while executing event handler")



from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='meross_iot',
    version='0.3.1.11',
    packages=find_packages(exclude=('tests',)),
    url='https://github.com/albertogeniola/MerossIot',
    license='MIT',
    author='Alberto Geniola',
    author_email='albertogeniola@gmail.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent'
    ],
    description='A simple library to deal with Meross devices. At the moment MSS110, MSS210, MSS310, MSS310H '
                'smart plugs and the MSS425E power strip. Other meross device might work out of the box with limited '
                'functionality. Give it a try and, in case of problems, let the developer know by opening an issue '
                'on Github.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='meross smartplug smartbulb iot mqtt domotic switch mss310 mss210 mss110 mss425e msl20 msg100',
    project_urls={
        'Documentation': 'https://github.com/albertogeniola/MerossIot',
        'Funding': 'https://donate.pypi.org',
        'Source': 'https://github.com/albertogeniola/MerossIot',
        'Tracker': 'https://github.com/albertogeniola/MerossIot/issues',
    },
    install_requires=[
        'paho-mqtt>=1.3.1',
        'requests>=2.19.1',
        'retrying>=1.3.3',
    ],
    python_requires='>=3.5',
    test_suite='tests',
    entry_points={
        'console_scripts': ['meross_info_gather=utilities.meross_info_gather:main']
    }
)
